;;;                      -*- Emacs-Lisp -*-
;;;
;;;                            mlh.el
;;;
;;;                 Copyright (C) 1992, 1993, 1994
;;;                         NIIBE Yutaka
;;;                  $B$K$$$Y(B $B$f$?$+(B ( g$B?7It(B $BM5(B )
;;;               e-mail: <gniibe@mri.co.jp>
;;;

;;;
;;; `mlh' means "Muchakucha" Loose "-na" "Henkan".
;;;
;;; It requires some kana-kanji conversion system when using Japanese.
;;;
;;; $B$3$N(B software $B$NG[I[!";HMQ$,$$$+$J$k7k2L$r0z$-5/$3$9$3(B
;;; $B$H$K$J$C$F$b(B NO WARRANTY $B$G$9!#(B
;;;
;;; $B$G$b!"J86g$r$$$&$N$O(B welcome $B$G$9!#$b$7$+$7$?$i!"<!$N(B
;;; version $B$G2~NI$5$l$k$+$bCN$l$^$;$s!#(B
;;;
;;; $B$H$O$$$C$F$b!"<!$N(B version $B$,=P$kJ]>Z$O$"$j$^$;$s!#$3$l(B
;;; $B$K$D$$$F$b(B NO WARRANTY $B$G$9!#(B;-)
;;;
;;;
;;; $BG[I[!";HMQ>r7o$O(B GNU General Public Licence version 2 $B$K4p$E$-$^$9!#(B
;;;
;;; $B$I$&$><+M3$KG[I[$7$F2<$5$$!#(B  $B$I$&$><+M3$K;H$C$F2<$5$$!#(B
;;; $B$G$b!"$$$i$J$$$H$$$&?M$KL5M}$K$"$2$k$N$O$d$a$^$7$g$&!#(B^_^
;;
;

;;
;;;(concat "\n;;; Version " mlh-version "\n;;; " (current-time-string))L/
;;;(concat "\n;;; Version " mlh-version "\n;;; " (current-time-string))L/
;;;
;;;
;; $B0J2<$O(B gN $B$N(B ~/.emacs $B$N(B mlh $B$K4X$9$kItJ,$G$9!#(B
;; $B9TF,$N(B ';> ' $B$O(B indicator $B$G$9!#(B

;> (setq mlh-comma-period-style 'comma-maru)
;> (load-library "mlh")
;> (load-library "mlh+wnn4-egg")
;> (define-key global-map " "  'mlh-space-bar-backward-henkan)
;> (setq mlh-userdef-table '(("maru" . "$B!;(B") ("kmaru" . "$B!|(B")
;>                           ("star" . "$B!y(B") ("kstar" . "$B!z(B")) )

(defconst mlh-version "0.5"
  "Version numbers of this version of mlh.")

(autoload 'mlh-zenkaku "mlh/nihongo" nil nil)
(autoload 'mlh-katakana "mlh/nihongo" nil nil)
(autoload 'mlh-hiragana "mlh/nihongo" nil nil)
(autoload 'mlh-hiragana-to-kanji "mlh/nihongo" nil nil)
(autoload 'mlh-kanji "mlh/nihongo" nil nil)
(autoload 'mlh-greek "mlh/greek" "Greek input for mlh." nil nil)
(autoload 'mlh-ltn "mlh/ltn" "Latin input for mlh." nil nil)
(autoload 'mlh-zhongwen "mlh/zhongwen" "zhongwen input for mlh." nil nil)
(autoload 'mlh-hangul "mlh/hangul" "Hangul input for mlh." nil nil)
;; added by M.Kuwada on 25-Mar-94
(autoload 'mlh-defrule-conditional "mlh-ext" "Conditional definition of conversion rules." nil nil)
(autoload 'mlh-defrule* "mlh-ext" "Redefinition of conversion rules." nil nil)
;; end of patch

(defvar mlh-cause-error-when-unknown-romaji-sequence
  "With t, mlh will cause an error, when it find
unknown romaji sequence.  Otherwise, it simply stop traversing."
  t)

(defvar henkan-region-function nil
  "Henkan region Function. ;-)")

(defun mlh-space-bar-backward-henkan ()
  "If the character preceding point is / (slash),
Do `mlh-backward-henkan'.  Then, invoke appropriate conversion, if needed.
Or else, execute command that space-bar invokes usually."
  (interactive)
  (let ((henkan-begin nil)
        (inhibit-henkan t))
    (mlh-backward-henkan)
    (if henkan-begin
        (if (or inhibit-henkan (= henkan-begin (point)))
            (egg:do-auto-fill)
          (let ((henkan-mode-map mlh-henkan-mode-map))
            (message "Converting...")
            (sit-for 0)
	    (funcall henkan-region-function henkan-begin (point))
            (message "") ))
      (setq this-command 'self-insert-command)
      (call-interactively 'self-insert-command))))

(defvar mlh-punctuations nil)
(if mlh-punctuations
    ()
  (setq mlh-punctuations "!()?;:"))

(defvar mlh-conversion-scheme-table
  '(
    (?- . mlh-kanji)
;    (?` . mlh-ltn)
;    (?' . mlh-ltn)
;    (?, . mlh-ltn)
    (?, . mlh-kanji)
    (?. . mlh-kanji)
;    (?^ . mlh-ltn)
;    (?~ . mlh-ltn)
;    (?\". mlh-ltn)
;    (?@ . mlh-ltn)
;    (?< . mlh-ltn)
;    (?> . mlh-ltn)
    (?a . mlh-kanji)
;    (?b . mlh-)
    (?c . mlh-capitalize)
    (?d . mlh-user-defined-conversion)
    (?e . mlh-kanji)
    (?f . mlh-hiragana)
    (?g . mlh-greek)
    (?h . mlh-hangul)
    (?i . mlh-kanji)
    (?j . mlh-jis-code)
    (?k . mlh-katakana)
;    (?l . mlh-ligature)
;    (?m . mlh-)
    (?n . mlh-kanji)
;    (?n . mlh-no-conversion)
    (?o . mlh-kanji)
    (?p . mlh-upcase-letter)
    (?q . mlh-quit)
;    (?r . mlh-)
    (?s . mlh-small-letter)
;    (?t . mlh-)
    (?u . mlh-kanji)
;    (?v . mlh-)
    (?w . mlh-white-space)
    (?x . mlh-execute)
;    (?y . mlh-)
    (?z . mlh-zhongwen)
    (?H . mlh-hiragana-to-kanji)
    (?L . mlh-lisp-expression)
    (?W . mlh-zenkaku-white)
    (?X . mlh-exit)
    (?Z . mlh-zenkaku)
))

(defun mlh-zenkaku-white ()
  (forward-char -1)
  (skip-chars-backward "0-9")
  (mlh-backward-henkan)
  (setq beg (point))
  (goto-char end-marker)
  (backward-delete-char 2)
  (let* ((str (buffer-substring beg (point)))
         (val (mlh-atoi str)))
    (delete-region beg (point))
    (if (= val 0)
        (setq val 1))
    (while (> val 0)
      (insert "$B!!(B")
      (setq val (1- val))))
  (if (null henkan-begin)
      (setq henkan-begin beg)))

(defun mlh-exit ()
  (goto-char end-marker)
  (backward-delete-char 2)
  (insert " ")
  (setq henkan-begin (point)))

(defun mlh-upcase-letter ()
  (forward-char -1)
  (skip-chars-backward "a-zA-Z0-9")
  (mlh-backward-henkan)
  (setq beg (point))
  (goto-char end-marker)
  (backward-delete-char 2)
  (upcase-region beg (point))
  (if (null henkan-begin)
      (setq henkan-begin beg)))

(defun mlh-capitalize ()
  (forward-char -1)
  (skip-chars-backward "a-zA-Z1-9")
  (mlh-backward-henkan)
  (setq beg (point))
  (goto-char end-marker)
  (backward-delete-char 2)
  (capitalize-region beg (point))
  (if (null henkan-begin)
      (setq henkan-begin beg)))

(defun mlh-jis-code ()
  (forward-char -1)
  (skip-chars-backward "0-9a-fA-F")
  (mlh-backward-henkan)
  (setq beg (point))
  (goto-char end-marker)
  (backward-delete-char 2)
  (let* ((str (buffer-substring beg (point)))
         (val (read-jis-code-from-string str)))
    ;;         ^--- this function is in egg.el
    (delete-region beg (point))
    (if val
        (insert (make-character lc-jp (car val) (cdr val)))
      (insert "?jis?")))
  (if (null henkan-begin)
      (setq henkan-begin beg)))

(defun mlh-lisp-expression ()
  (forward-char -1)
  (let ((stab (syntax-table)))
    (unwind-protect
        (progn
          (set-syntax-table emacs-lisp-mode-syntax-table)
          (forward-sexp -1))
      (set-syntax-table stab)))
  (mlh-backward-henkan)
  (setq beg (point))
  (goto-char end-marker)
  (backward-delete-char 2)
  (let* ((exp-str
          (buffer-substring beg (point)))
         (exp (car (read-from-string exp-str)))
         (result (eval exp)))
    (delete-region beg (point))
    (insert (format "%s" result)))
  (if (null henkan-begin)
      (setq henkan-begin beg)))

(defun mlh-quit ()
  (goto-char end-marker)
  (backward-delete-char 2)
  (setq henkan-begin (point)))
  
(defun mlh-no-conversion ()
  (forward-char -1)
  (skip-chars-backward "\041-\056\060-\176")
  (mlh-backward-henkan)
  (setq beg (point))
  (goto-char end-marker)
  (backward-delete-char 2)
  (if (null henkan-begin)
      (setq henkan-begin beg)))

(fset 'mlh-small-letter (symbol-function 'mlh-no-conversion))

(defun mlh-white-space ()
  (forward-char -1)
  (skip-chars-backward "0-9")
  (mlh-backward-henkan)
  (setq beg (point))
  (goto-char end-marker)
  (backward-delete-char 2)
  (let* ((str (buffer-substring beg (point)))
         (val (mlh-atoi str)))
    (delete-region beg (point))
    (if (= val 0)
        (setq val 1))
    (insert (make-string val ?\ )))
  (if (null henkan-begin)
      (setq henkan-begin beg)))

(defun mlh-execute ()
  (forward-char -1)
  (if (fboundp 'mlh-userdef-function)
      (mlh-userdef-function)
    (mlh-backward-henkan)
    (setq beg (point))
    (goto-char end-marker)
    (backward-delete-char 2)
    (if (null henkan-begin)
        (setq henkan-begin beg))))

(defun mlh-backward-henkan ()
  "For each words seperated by / (slash), do conversion.
Accoding to a character preceding slash, conversion scheme are selected.

CHAR.  MNEMONIC             CONVERSION SCHEME

  H    Hiragana to kanji    Convert Hiragana to Kanji
  L    Lisp                 Evaluate as Emacs-Lisp expression
  W    zenkaku White space  Insert Zenkaku spaces
  X    eXit                 Quit going backward, insert space
  Z    Zenkaku              Convert to Zenkaku
  c    Capitalize           Capitalize
  d    user Definition      Convert with user definition table
  f    Firagana ??          Convert to Hiragana
  g    Greek letter         Convert to single greek letter
  h    Hangul               Convert to Hangul
  j    Jis-code             Convert to character which has code
  k    Katakana             Convert to Katakana
  l    Ligature             Ligature (not implemented yet)
  p    uPcase letter        uPcase
  q    Quit                 Quit going backward
  s    Small letter         No conversion
  w    White space          Insert spaces
  x    eXecute              Call user defined function
  z    Zhongwen             Convert to Zhongwen
    OTHERWISE               Convert to KANJI
"
  (if (eq (preceding-char) ?/)
      (let ((end-marker (point-marker))
            (char nil)
            (beg nil))
        (set-marker-type end-marker t)
        (unwind-protect
            (let (scheme)
              (backward-char 1)
              (setq char (preceding-char))
              (cond 
               ((setq scheme (assq char mlh-conversion-scheme-table))
                (funcall (cdr scheme)))
               (t
                (goto-char end-marker)))
              (if beg
                  (progn
                    (goto-char beg)
                    (mlh-do-spacing)
                    (goto-char end-marker))))
          (set-marker end-marker nil)))))


;;; romaji to kana conversion
;;; using automata

;; <state>::= ( <output> . <map> )
;; <output>::= nil | <string>
;; <map>::= alist of (<ch> . <state>)

(defvar mlh-start-state nil)
(defvar mlh-set-of-chars nil)

(defvar mlh-start-state-sym nil)
(defvar mlh-set-of-chars-sym nil)

(defmacro mlh-define-automata (start-state set-of-chars)
  (list
   'progn
   (list 'setq 'mlh-start-state-sym (list 'quote start-state))
   (list 'setq 'mlh-set-of-chars-sym (list 'quote set-of-chars))
   (list 'setq 'mlh-start-state '(cons nil nil))
   (list 'setq 'mlh-set-of-chars  nil)))

(defun mlh-end-of-definition ()
  (set mlh-start-state-sym mlh-start-state)
  (set mlh-set-of-chars-sym mlh-set-of-chars))

(defun mlh-defrule (seq output)
  (let ((state mlh-start-state)
        (i (1- (length seq))))
    (while (>= i 0)
      (let* ((ch (aref seq i))
             (next-state (mlh-next-state state ch)))
        (if (null (memq ch mlh-set-of-chars))
            (setq mlh-set-of-chars (cons ch mlh-set-of-chars)))
        (if next-state
            (setq state next-state)
          (setq next-state (cons nil nil)) ; create new state
          (setcdr state (cons (cons ch next-state) (cdr state)))
          (setq state next-state)))
      (setq i (1- i)))
    (if (car state)
        (error "Ambiguous rule")
      (setcar state output))))

(defun mlh-next-state (state ch)
  (cdr (assq ch (cdr state))))
  
(defun mlh-roma-kana-backward (start-state set-of-chars)
  "Traverse backward and return corresponding converted string.
START-STATE is the start state of the automaton.
SET-OF-CHARS is the set of valid characters of the automaton.
"
  (let ((start (point))
        (end nil)
        (output nil)
        (state start-state)
        (result "")
        (ch nil)
        (in-progress t))
    (while in-progress
      (if (car state) ;; Is it final state?
          (setq end (point)
                output (car state)))
      (if (or (bobp)
              (progn
                (setq ch (preceding-char))
                (not (memq ch set-of-chars))))
          (let ((point (point)))
            (setq in-progress nil)
            (if end
                (progn
                  (goto-char end)
                  (setq result (concat output result)))
              (goto-char start))
            (if (and mlh-cause-error-when-unknown-romaji-sequence
                     (/= point (point)) )
                    (error "Unknown roma-ji sequence.") ))
        (setq state (mlh-next-state state ch))
        (if state
            (forward-char -1)
          (if end
              (progn
                (goto-char end)
                (setq result (concat output result)
                      start end
                      end nil
                      output nil
                      state start-state))
            (goto-char start)
            (if mlh-cause-error-when-unknown-romaji-sequence
                (error "Unknown roma-ji sequence.")
              (setq in-progress nil))))))
    result))


(defun mlh-atoi (str)
  "Ascii string to Integer."
  (let ((val 0)
        (len (length str))
        (i 0))
    (while (< i len)
      (setq val (+ (* val 10) (- (aref str i) ?0))
            i (1+ i)))
    val))

(defvar mlh-syntax-table nil
  "Syntax table of mlh, which are used to determine spacing.")
(if mlh-syntax-table
    ()
  (setq mlh-syntax-table (copy-syntax-table emacs-lisp-mode-syntax-table))
  (modify-syntax-entry ?! "." mlh-syntax-table)
  (modify-syntax-entry ?$ "'" mlh-syntax-table)
  (modify-syntax-entry ?% "'" mlh-syntax-table)
  (modify-syntax-entry ?& "'" mlh-syntax-table)
  (modify-syntax-entry ?{ "(}" mlh-syntax-table)
  (modify-syntax-entry ?} "){" mlh-syntax-table)
)

(defvar mlh-space-control nil
  "Alist that determines inserting space.")
(if mlh-space-control
    ()
  (setq mlh-space-control
  '(
    (("a".?w).("a".?w))
    (("a".?w).("a".?_))
    (("a".?w).("Hj".?e))
    (("a".?w).("Cj".?e))
    (("a".?_).("a".?w))
    (("a".?_).("a".?_))
    (("a".?_).("Hj".?e))
    (("a".?_).("Cj".?e))
    (("a".?.).("a".?w))
    (("a".?.).("a".?_))
    (("a".?_).("Hj".?e))
    (("a".?_).("Cj".?e))
    (("Hj".?e).("a".?w))
    (("Cj".?e).("a".?w))
    (("Hj".?e).("a".?_))
    (("Cj".?e).("a".?_))
    )))

(defun mlh-do-spacing ()
  "Arrange spacing as you like."
  (if (bobp)
      ()
    (let ((s-tab (syntax-table))
          s-pc s-fc
          c-pc c-fc)
      (unwind-protect
          (progn
            (set-syntax-table mlh-syntax-table)
            (setq s-pc (char-syntax (preceding-char))
                  s-fc (char-syntax (following-char))))
        (set-syntax-table s-tab))
      (setq c-pc (char-category (preceding-char))
            c-fc (char-category (following-char)))
      (if (member (cons (cons c-pc s-pc) (cons c-fc s-fc)) mlh-space-control)
          (progn
            (and henkan-begin
                 (>= henkan-begin (point))
                 (setq henkan-begin (1+ henkan-begin)))
            (insert " "))))))

(defun member (elt list)
  "Returns non-nil if ELT is an element of LIST.  Comparison done with EQUAL.
The value is actually the tail of LIST whose car is ELT."
  (let ((in-search t)
        (result nil))
    (while (and in-search list)
      (cond ((equal elt (car list))
             (setq in-search nil)
             (setq result list)))
      (setq list (cdr list)))
    result))

(defvar mlh-select-mode-map (make-keymap))

;;; acutually this map is not necessary now. for future extention
(defvar mlh-select-mode-esc-map (make-keymap))

(let ((ch 0))
  (while (<= ch 127)
    (define-key mlh-select-mode-map (char-to-string ch) 'undefined)
    (define-key mlh-select-mode-esc-map (char-to-string ch) 'undefined)
    (setq ch (1+ ch))))

(let ((ch 32))
  (while (< ch 127)
    (define-key mlh-select-mode-map (char-to-string ch)
      'mlh-select-kakutei-and-self-insert)
    (setq ch (1+ ch))))

(define-key mlh-select-mode-map "\C-m" 'mlh-select-kakutei-and-self-insert)
(define-key mlh-select-mode-map "\C-b" 'mlh-select-prev-candidate)
(define-key mlh-select-mode-map "\C-f" 'mlh-select-next-candidate)
(define-key mlh-select-mode-map "\177" 'mlh-select-prev-candidate)
(define-key mlh-select-mode-map " " 'mlh-select-next-candidate)
(define-key mlh-select-mode-map "/" 'mlh-select-kakutei)

(defun mlh-select-insert-candidate (n)
  (delete-region beg (point))
  (insert (nth n candidates)))

(defun mlh-select-prev-candidate ()
  (interactive)
  (setq current-candidate (1- current-candidate))
  (if (< current-candidate 0)
      (setq current-candidate (1- number-of-candidates)))
  (mlh-select-insert-candidate current-candidate))

(defun mlh-select-next-candidate ()
  (interactive)
  (setq current-candidate (1+ current-candidate))
  (if (>= current-candidate number-of-candidates)
      (setq current-candidate 0))
  (mlh-select-insert-candidate current-candidate))

(defun mlh-recursive-edit-select (beg end candidates)
  (mlh-select-insert-candidate 0)
  (and (boundp 'disable-undo) (setq disable-undo t))
  (let ((old-local-map (current-local-map))
	(number-of-candidates (length candidates))
	(current-candidate 0))
    (use-local-map mlh-select-mode-map)
    (recursive-edit)
    (use-local-map old-local-map)))

(defun mlh-select-kakutei-and-self-insert ()
  (interactive)
  (setq unread-command-char last-command-char)
  (mlh-select-kakutei))

(defun mlh-select-kakutei ()
  (interactive)
  (and (boundp 'disable-undo) (setq disable-undo nil))
  (exit-recursive-edit))

(defun mlh-user-defined-conversion ()
  (forward-char -1)
  (skip-chars-backward "-a-zA-Z")
  (mlh-backward-henkan)
  (setq beg (point))
  (goto-char end-marker)
  (backward-delete-char 2)
  (let* ((str (buffer-substring beg (point)))
         (userdef (mlh-userdef<==string str)))
    (cond ((stringp userdef)
	   (delete-region beg (point))
	   (insert userdef))
	  ((null userdef)
	   (delete-region beg (point))
	   ;; (add-userdef) (insert new-userdef)
	   (insert "?udef?"))
	  ((consp userdef)
	   (mlh-recursive-edit-select beg (point) userdef))))
  (if (null henkan-begin)
      (setq henkan-begin beg)))

(defvar mlh-userdef-table nil
  "Convertion table of words(string) to another words(string).")

(defun mlh-userdef<==string (str)
  "Convert string to another string with `mlh-userdef-table'"
  (cdr (assoc str mlh-userdef-table)))
