(defvar mlh-kanji-function 'mlh-kanji-with-henkan-region-function)

(defun mlh-kanji ()
  (funcall mlh-kanji-function))

(defun mlh-kanji-with-henkan-region-function ()
  (let ((string (mlh-roma-kana-backward
                 mlh-hira-start-state mlh-hira-chars)))
    (mlh-backward-henkan)
    (setq inhibit-henkan nil)
    (setq beg (point))
    (goto-char end-marker)
    (delete-region beg (point))
    (insert string))
  (if (null henkan-begin)
      (setq henkan-begin beg)))

(defun mlh-hiragana ()
  (forward-char -1)
  (let ((string (mlh-roma-kana-backward
                 mlh-hira-start-state mlh-hira-chars)))
    (mlh-backward-henkan)
    (setq beg (point))
    (goto-char end-marker)
    (delete-region beg (point))
    (insert string))
  (setq henkan-begin (point)))

(defun mlh-hiragana-to-kanji ()
  (forward-char -1)
  (skip-chars-backward "$B$!(B-$B$s(B")
  (mlh-backward-henkan)
  (setq beg (point))
  (setq inhibit-henkan nil)
  (goto-char end-marker)
  (backward-delete-char 2)
  (if (null henkan-begin)
      (setq henkan-begin beg)))

(defun mlh-katakana ()
  (forward-char -1)
  (let ((string (mlh-roma-kana-backward
                 mlh-hira-start-state mlh-hira-chars)))
    (mlh-backward-henkan)
    (setq beg (point))
    (goto-char end-marker)
    (delete-region beg (point))
    (insert (mlh-hira-to-kata string)))
  (if (null henkan-begin)
      (setq henkan-begin beg)))

(defun mlh-zenkaku ()
  (forward-char -1)
  (skip-chars-backward "\041-\056\060-\176")
  (mlh-backward-henkan)
  (setq beg (point))
  (goto-char end-marker)
  (backward-delete-char 2)
  (zenkaku-region beg (point))
  (if (null henkan-begin)
      (setq henkan-begin beg)))

(defun mlh-hira-to-kata (str)
  "Convert hiragana to katakana in STR."
  (let ((result (copy-sequence str))
        (i 0))
    (while (setq i (string-match "[$B$!(B-$B$s(B]" str i))
      (aset result (1+ i) ?\245)
      (setq i (+ i 3)))
    result))

(defvar mlh-comma-period-style nil
  "Variable which effects style of comma and period.
Possible value is:
       comma-maru        ,     $B!#(B (default)
       comma2-maru       $B!$(B    $B!#(B
       comma-period      ,     .
       comma-period2     ,     $B!%(B
       comma2-period     $B!$(B    .
       comma2-period2    $B!$(B    $B!%(B
       ten-period2       $B!"(B    $B!%(B
       ten-maru          $B!"(B    $B!#(B
")

;;; Produce Automaton for Roma-ji to Hiragana conversion.
;;; Coded by NIIBE Yutaka <gniibe@mri.co.jp>
;;;
;;; The codes in this file are based on egg-roma-kana.el, which is,
;;; distributed in Nemacs 3.3.2
;; Basic Roma-to-Kana Translation Table for Egg
;; Coded by S.Tomura, Electrotechnical Lab. (tomura@etl.go.jp)
;;;;

(mlh-define-automata mlh-hira-start-state mlh-hira-chars)

(mlh-defrule "0" "0")
(mlh-defrule "1" "1")
(mlh-defrule "2" "2")
(mlh-defrule "3" "3")
(mlh-defrule "4" "4")
(mlh-defrule "5" "5")
(mlh-defrule "6" "6")
(mlh-defrule "7" "7")
(mlh-defrule "8" "8")
(mlh-defrule "9" "9")

(mlh-defrule "..."  "...")
(cond ((eq mlh-comma-period-style 'comma-period)
       (mlh-defrule ","    ", ")
       (mlh-defrule "."    ". ")
       )
      ((eq mlh-comma-period-style 'comma2-maru)
       (mlh-defrule ","    "$B!$(B")
       (mlh-defrule "."    "$B!#(B")
       )
      ((eq mlh-comma-period-style 'comma-period2)
       (mlh-defrule ","    ", ")
       (mlh-defrule "."    "$B!%(B")
       )
      ((eq mlh-comma-period-style 'comma2-period)
       (mlh-defrule ","    "$B!$(B")
       (mlh-defrule "."    ". ")
       )
      ((eq mlh-comma-period-style 'comma2-period2)
       (mlh-defrule ","    "$B!$(B")
       (mlh-defrule "."    "$B!%(B")
       )
      ((eq mlh-comma-period-style 'ten-period2)
       (mlh-defrule ","    "$B!"(B")
       (mlh-defrule "."    "$B!%(B")
       )
      ((eq mlh-comma-period-style 'ten-maru)
       (mlh-defrule ","    "$B!"(B")
       (mlh-defrule "."    "$B!#(B")
       )
      (t			    ; (eq mlh-comma-period-style 'comma-maru)
       (mlh-defrule ","    ", ")
       (mlh-defrule "."    "$B!#(B")))

(mlh-defrule "-"    "$B!<(B")
(mlh-defrule "["    "$B!V(B")
(mlh-defrule "]"    "$B!W(B")
(mlh-defrule "xa"   "$B$!(B")
(mlh-defrule "a"    "$B$"(B")
(mlh-defrule "xi"   "$B$#(B")
(mlh-defrule "i"    "$B$$(B")
(mlh-defrule "xu"   "$B$%(B")
(mlh-defrule "u"    "$B$&(B")
(mlh-defrule "xwi"  "$B$&$#(B")
(mlh-defrule "xwe"  "$B$&$'(B")
(mlh-defrule "xwo"  "$B$&$)(B")
(mlh-defrule "xe"   "$B$'(B")
(mlh-defrule "e"    "$B$((B")
(mlh-defrule "xo"   "$B$)(B")
(mlh-defrule "o"    "$B$*(B")
(mlh-defrule "ka"   "$B$+(B")
(mlh-defrule "ga"   "$B$,(B")
(mlh-defrule "gya"   "$B$.$c(B")
(mlh-defrule "ki"   "$B$-(B")
(mlh-defrule "kya"  "$B$-$c(B")
(mlh-defrule "kyu"  "$B$-$e(B")
(mlh-defrule "kyo"  "$B$-$g(B")
(mlh-defrule "gi"   "$B$.(B")
(mlh-defrule "ku"   "$B$/(B")
(mlh-defrule "kwu"  "$B$/(B")
(mlh-defrule "kwi"  "$B$/$#(B")
(mlh-defrule "kwe"  "$B$/$'(B")
(mlh-defrule "kwo"  "$B$/$)(B")
(mlh-defrule "kwa"  "$B$/$n(B")
(mlh-defrule "gu"   "$B$0(B")
(mlh-defrule "gwu"  "$B$0(B")
(mlh-defrule "gwi"  "$B$0$#(B")
(mlh-defrule "gwe"  "$B$0$'(B")
(mlh-defrule "gwo"  "$B$0$)(B")
(mlh-defrule "gwa"  "$B$0$n(B")
(mlh-defrule "gyu"   "$B$.$e(B")
(mlh-defrule "ke"   "$B$1(B")
(mlh-defrule "ge"   "$B$2(B")
(mlh-defrule "ko"   "$B$3(B")
(mlh-defrule "go"   "$B$4(B")
(mlh-defrule "gyo"   "$B$.$g(B")
(mlh-defrule "sa"   "$B$5(B")
(mlh-defrule "za"   "$B$6(B")
(mlh-defrule "shi"  "$B$7(B")
(mlh-defrule "si"   "$B$7(B")
(mlh-defrule "she"  "$B$7$'(B")
(mlh-defrule "sha"  "$B$7$c(B")
(mlh-defrule "sya"  "$B$7$c(B")
(mlh-defrule "shu"  "$B$7$e(B")
(mlh-defrule "syu"  "$B$7$e(B")
(mlh-defrule "sho"  "$B$7$g(B")
(mlh-defrule "syo"  "$B$7$g(B")
(mlh-defrule "ji"   "$B$8(B")
(mlh-defrule "zi"   "$B$8(B")
(mlh-defrule "je"   "$B$8$'(B")
(mlh-defrule "ja"   "$B$8$c(B")
(mlh-defrule "zya"  "$B$8$c(B")
(mlh-defrule "ju"   "$B$8$e(B")
(mlh-defrule "zyu"  "$B$8$e(B")
(mlh-defrule "jo"   "$B$8$g(B")
(mlh-defrule "zyo"  "$B$8$g(B")
(mlh-defrule "su"   "$B$9(B")
(mlh-defrule "zu"   "$B$:(B")
(mlh-defrule "se"   "$B$;(B")
(mlh-defrule "ze"   "$B$<(B")
(mlh-defrule "so"   "$B$=(B")
(mlh-defrule "zo"   "$B$>(B")
(mlh-defrule "ta"   "$B$?(B")
(mlh-defrule "da"   "$B$@(B")
(mlh-defrule "chi"  "$B$A(B")
(mlh-defrule "ti"   "$B$A(B")
(mlh-defrule "che"  "$B$A$'(B")
(mlh-defrule "cha"  "$B$A$c(B")
(mlh-defrule "tya"  "$B$A$c(B")
(mlh-defrule "chu"  "$B$A$e(B")
(mlh-defrule "tyu"  "$B$A$e(B")
(mlh-defrule "cho"  "$B$A$g(B")
(mlh-defrule "tyo"  "$B$A$g(B")
(mlh-defrule "di"   "$B$B(B")
(mlh-defrule "ddye" "$B$B$'(B")
(mlh-defrule "dye"  "$B$B$'(B")
(mlh-defrule "ddya" "$B$B$c(B")
(mlh-defrule "dya"  "$B$B$c(B")
(mlh-defrule "ddyu" "$B$B$e(B")
(mlh-defrule "dyu"  "$B$B$e(B")
(mlh-defrule "ddyo" "$B$B$g(B")
(mlh-defrule "dyo"  "$B$B$g(B")
(mlh-defrule "xtsu" "$B$C(B")
(mlh-defrule "xtu"  "$B$C(B")
(mlh-defrule "kka"  "$B$C$+(B")
(mlh-defrule "gga"  "$B$C$,(B")
(mlh-defrule "kki"  "$B$C$-(B")
(mlh-defrule "kkya" "$B$C$-$c(B")
(mlh-defrule "kkyu" "$B$C$-$e(B")
(mlh-defrule "kkyo" "$B$C$-$g(B")
(mlh-defrule "ggi"  "$B$C$.(B")
(mlh-defrule "kku"  "$B$C$/(B")
(mlh-defrule "kkwu" "$B$C$/(B")
(mlh-defrule "kkwi" "$B$C$/$#(B")
(mlh-defrule "kkwe" "$B$C$/$'(B")
(mlh-defrule "kkwo" "$B$C$/$)(B")
(mlh-defrule "kkwa" "$B$C$/$n(B")
(mlh-defrule "ggu"  "$B$C$0(B")
(mlh-defrule "ggwu" "$B$C$0(B")
(mlh-defrule "ggwi" "$B$C$0$#(B")
(mlh-defrule "ggwe" "$B$C$0$'(B")
(mlh-defrule "ggwo" "$B$C$0$)(B")
(mlh-defrule "ggwa" "$B$C$0$n(B")
(mlh-defrule "kke"  "$B$C$1(B")
(mlh-defrule "gge"  "$B$C$2(B")
(mlh-defrule "kko"  "$B$C$3(B")
(mlh-defrule "ggo"  "$B$C$4(B")
(mlh-defrule "ssa"  "$B$C$5(B")
(mlh-defrule "zza"  "$B$C$6(B")
(mlh-defrule "sshi" "$B$C$7(B")
(mlh-defrule "ssi"  "$B$C$7(B")
(mlh-defrule "sshe" "$B$C$7$'(B")
(mlh-defrule "ssha" "$B$C$7$c(B")
(mlh-defrule "ssya" "$B$C$7$c(B")
(mlh-defrule "sshu" "$B$C$7$e(B")
(mlh-defrule "ssyu" "$B$C$7$e(B")
(mlh-defrule "ssho" "$B$C$7$g(B")
(mlh-defrule "ssyo" "$B$C$7$g(B")
(mlh-defrule "jji"  "$B$C$8(B")
(mlh-defrule "zzi"  "$B$C$8(B")
(mlh-defrule "jje"  "$B$C$8$'(B")
(mlh-defrule "jja"  "$B$C$8$c(B")
(mlh-defrule "zzya" "$B$C$8$c(B")
(mlh-defrule "jju"  "$B$C$8$e(B")
(mlh-defrule "zzyu" "$B$C$8$e(B")
(mlh-defrule "jjo"  "$B$C$8$g(B")
(mlh-defrule "zzyo" "$B$C$8$g(B")
(mlh-defrule "ssu"  "$B$C$9(B")
(mlh-defrule "zzu"  "$B$C$:(B")
(mlh-defrule "sse"  "$B$C$;(B")
(mlh-defrule "zze"  "$B$C$<(B")
(mlh-defrule "sso"  "$B$C$=(B")
(mlh-defrule "zzo"  "$B$C$>(B")
(mlh-defrule "tta"  "$B$C$?(B")
(mlh-defrule "dda"  "$B$C$@(B")
(mlh-defrule "cchi" "$B$C$A(B")
(mlh-defrule "tchi" "$B$C$A(B")
(mlh-defrule "tti"  "$B$C$A(B")
(mlh-defrule "cche" "$B$C$A$'(B")
(mlh-defrule "tche" "$B$C$A$'(B")
(mlh-defrule "ccha" "$B$C$A$c(B")
(mlh-defrule "tcha" "$B$C$A$c(B")
(mlh-defrule "ttya" "$B$C$A$c(B")
(mlh-defrule "cchu" "$B$C$A$e(B")
(mlh-defrule "tchu" "$B$C$A$e(B")
(mlh-defrule "ttyu" "$B$C$A$e(B")
(mlh-defrule "ccho" "$B$C$A$g(B")
(mlh-defrule "tcho" "$B$C$A$g(B")
(mlh-defrule "ttyo" "$B$C$A$g(B")
(mlh-defrule "ddi"  "$B$C$B(B")
(mlh-defrule "ttsu" "$B$C$D(B")
(mlh-defrule "ttu"  "$B$C$D(B")
(mlh-defrule "ttsa" "$B$C$D$!(B")
(mlh-defrule "ttsi" "$B$C$D$#(B")
(mlh-defrule "ttse" "$B$C$D$'(B")
(mlh-defrule "ttso" "$B$C$D$)(B")
(mlh-defrule "ddu"  "$B$C$E(B")
(mlh-defrule "tte"  "$B$C$F(B")
(mlh-defrule "dde"  "$B$C$G(B")
(mlh-defrule "tto"  "$B$C$H(B")
(mlh-defrule "ddo"  "$B$C$I(B")
(mlh-defrule "hha"  "$B$C$O(B")
(mlh-defrule "bba"  "$B$C$P(B")
(mlh-defrule "ppa"  "$B$C$Q(B")
(mlh-defrule "hhi"  "$B$C$R(B")
(mlh-defrule "hhya" "$B$C$R$c(B")
(mlh-defrule "hhyu" "$B$C$R$e(B")
(mlh-defrule "hhyo" "$B$C$R$g(B")
(mlh-defrule "bbi"  "$B$C$S(B")
(mlh-defrule "bbya" "$B$C$S$c(B")
(mlh-defrule "bbyu" "$B$C$S$e(B")
(mlh-defrule "bbyo" "$B$C$S$g(B")
(mlh-defrule "ppi"  "$B$C$T(B")
(mlh-defrule "ppya" "$B$C$T$c(B")
(mlh-defrule "ppyu" "$B$C$T$e(B")
(mlh-defrule "ppyo" "$B$C$T$g(B")
(mlh-defrule "ffu"  "$B$C$U(B")
(mlh-defrule "hhu"  "$B$C$U(B")
(mlh-defrule "ffa"  "$B$C$U$!(B")
(mlh-defrule "ffi"  "$B$C$U$#(B")
(mlh-defrule "ffe"  "$B$C$U$'(B")
(mlh-defrule "ffo"  "$B$C$U$)(B")
(mlh-defrule "bbu"  "$B$C$V(B")
(mlh-defrule "ppu"  "$B$C$W(B")
(mlh-defrule "hhe"  "$B$C$X(B")
(mlh-defrule "bbe"  "$B$C$Y(B")
(mlh-defrule "ppe"  "$B$C$Z(B")
(mlh-defrule "hho"  "$B$C$[(B")
(mlh-defrule "bbo"  "$B$C$\(B")
(mlh-defrule "ppo"  "$B$C$](B")
(mlh-defrule "mmya" "$B$C$_$c(B")
(mlh-defrule "mmyu" "$B$C$_$e(B")
(mlh-defrule "mmyo" "$B$C$_$g(B")
(mlh-defrule "lla"  "$B$C$i(B")
(mlh-defrule "rra"  "$B$C$i(B")
(mlh-defrule "lli"  "$B$C$j(B")
(mlh-defrule "rri"  "$B$C$j(B")
(mlh-defrule "llya" "$B$C$j$c(B")
(mlh-defrule "trya" "$B$C$j$c(B")
(mlh-defrule "llyu" "$B$C$j$e(B")
(mlh-defrule "rryu" "$B$C$j$e(B")
(mlh-defrule "llyo" "$B$C$j$g(B")
(mlh-defrule "rryo" "$B$C$j$g(B")
(mlh-defrule "llu"  "$B$C$k(B")
(mlh-defrule "rru"  "$B$C$k(B")
(mlh-defrule "lle"  "$B$C$l(B")
(mlh-defrule "rre"  "$B$C$l(B")
(mlh-defrule "llo"  "$B$C$m(B")
(mlh-defrule "rro"  "$B$C$m(B")
(mlh-defrule "vvu"  "$B$C%t(B")
(mlh-defrule "vva"  "$B$C%t$!(B")
(mlh-defrule "vvi"  "$B$C%t$#(B")
(mlh-defrule "vve"  "$B$C%t$'(B")
(mlh-defrule "vvo"  "$B$C%t$)(B")
(mlh-defrule "tsu"  "$B$D(B")
(mlh-defrule "tu"   "$B$D(B")
(mlh-defrule "tsa"  "$B$D$!(B")
(mlh-defrule "tsi"  "$B$D$#(B")
(mlh-defrule "tse"  "$B$D$'(B")
(mlh-defrule "tso"  "$B$D$)(B")
(mlh-defrule "du"   "$B$E(B")
(mlh-defrule "te"   "$B$F(B")
(mlh-defrule "xti"  "$B$F$#(B")
(mlh-defrule "de"   "$B$G(B")
(mlh-defrule "ddyi" "$B$G$#(B")
(mlh-defrule "dyi"  "$B$G$#(B")
(mlh-defrule "xdi"  "$B$G$#(B")
(mlh-defrule "xde"  "$B$G$'(B")
(mlh-defrule "to"   "$B$H(B")
(mlh-defrule "do"   "$B$I(B")
(mlh-defrule "xdu"  "$B$I$%(B")
(mlh-defrule "xdo"  "$B$I$)(B")
(mlh-defrule "na"   "$B$J(B")
(mlh-defrule "ni"   "$B$K(B")
(mlh-defrule "nya"  "$B$K$c(B")
(mlh-defrule "nyu"  "$B$K$e(B")
(mlh-defrule "nyo"  "$B$K$g(B")
(mlh-defrule "nu"   "$B$L(B")
(mlh-defrule "ne"   "$B$M(B")
(mlh-defrule "no"   "$B$N(B")
(mlh-defrule "ha"   "$B$O(B")
(mlh-defrule "ba"   "$B$P(B")
(mlh-defrule "pa"   "$B$Q(B")
(mlh-defrule "hi"   "$B$R(B")
(mlh-defrule "hya"  "$B$R$c(B")
(mlh-defrule "hyu"  "$B$R$e(B")
(mlh-defrule "hyo"  "$B$R$g(B")
(mlh-defrule "bi"   "$B$S(B")
(mlh-defrule "bya"  "$B$S$c(B")
(mlh-defrule "byu"  "$B$S$e(B")
(mlh-defrule "byo"  "$B$S$g(B")
(mlh-defrule "pi"   "$B$T(B")
(mlh-defrule "pya"  "$B$T$c(B")
(mlh-defrule "pyu"  "$B$T$e(B")
(mlh-defrule "pyo"  "$B$T$g(B")
(mlh-defrule "fu"   "$B$U(B")
(mlh-defrule "hu"   "$B$U(B")
(mlh-defrule "fa"   "$B$U$!(B")
(mlh-defrule "fi"   "$B$U$#(B")
(mlh-defrule "fe"   "$B$U$'(B")
(mlh-defrule "fo"   "$B$U$)(B")
(mlh-defrule "bu"   "$B$V(B")
(mlh-defrule "pu"   "$B$W(B")
(mlh-defrule "he"   "$B$X(B")
(mlh-defrule "be"   "$B$Y(B")
(mlh-defrule "pe"   "$B$Z(B")
(mlh-defrule "ho"   "$B$[(B")
(mlh-defrule "bo"   "$B$\(B")
(mlh-defrule "po"   "$B$](B")
(mlh-defrule "ma"   "$B$^(B")
(mlh-defrule "mi"   "$B$_(B")
(mlh-defrule "mya"  "$B$_$c(B")
(mlh-defrule "myu"  "$B$_$e(B")
(mlh-defrule "myo"  "$B$_$g(B")
(mlh-defrule "mu"   "$B$`(B")
(mlh-defrule "me"   "$B$a(B")
(mlh-defrule "mo"   "$B$b(B")
(mlh-defrule "xya"  "$B$c(B")
(mlh-defrule "ya"   "$B$d(B")
(mlh-defrule "xyu"  "$B$e(B")
(mlh-defrule "yu"   "$B$f(B")
(mlh-defrule "xyo"  "$B$g(B")
(mlh-defrule "yo"   "$B$h(B")
(mlh-defrule "la"   "$B$i(B")
(mlh-defrule "ra"   "$B$i(B")
(mlh-defrule "li"   "$B$j(B")
(mlh-defrule "ri"   "$B$j(B")
(mlh-defrule "lya"  "$B$j$c(B")
(mlh-defrule "rya"  "$B$j$c(B")
(mlh-defrule "lyu"  "$B$j$e(B")
(mlh-defrule "ryu"  "$B$j$e(B")
(mlh-defrule "lyo"  "$B$j$g(B")
(mlh-defrule "ryo"  "$B$j$g(B")
(mlh-defrule "lu"   "$B$k(B")
(mlh-defrule "ru"   "$B$k(B")
(mlh-defrule "le"   "$B$l(B")
(mlh-defrule "re"   "$B$l(B")
(mlh-defrule "lo"   "$B$m(B")
(mlh-defrule "ro"   "$B$m(B")
(mlh-defrule "xwa"  "$B$n(B")
(mlh-defrule "wa"   "$B$o(B")
(mlh-defrule "wi"   "$B$p(B")
(mlh-defrule "we"   "$B$q(B")
(mlh-defrule "wo"   "$B$r(B")
(mlh-defrule "N"    "$B$s(B")
(mlh-defrule "n"    "$B$s(B")
(mlh-defrule "n'"   "$B$s(B")
(mlh-defrule "vu"   "$B%t(B")
(mlh-defrule "va"   "$B%t$!(B")
(mlh-defrule "vi"   "$B%t$#(B")
(mlh-defrule "ve"   "$B%t$'(B")
(mlh-defrule "vo"   "$B%t$)(B")
(mlh-defrule "xka"  "$B%u(B")
(mlh-defrule "xke"  "$B%v(B")

(mlh-end-of-definition)
